@@ -1,17 +1,19 @@
 import gspread
 from gspread.utils import rowcol_to_a1, a1_to_rowcol, ValueRenderOption, ValueInputOption
+from gspread.auth import local_server_flow
+from gspread.auth import DEFAULT_SCOPES, DEFAULT_CREDENTIALS_FILENAME, DEFAULT_AUTHORIZED_USER_FILENAME
 from .gspreadsheet_retry import SpreadsheetRetry, WorksheetRetry, ClientRetry
 from .gspreadsheet_retry import exceptions, retry, error_quota_req
 from .format_cell import CellFormat, ColorMap
 import logging
 from re import compile, IGNORECASE
-from os import getenv
+from os import getenv, unlink, path
 
 """
 GoogleSheets HighLevel wrapper around gspread
 """
 
-__version__ = "1.0.1"
+__version__ = "1.0.2"
 
 logger = logging.getLogger('GoogleSheets')
 
@@ -58,6 +60,30 @@ class GridIndex(object):
     def __eq__(self, other):
         return  self.start == other.start and self.end == other.end
 
+class DataCache(object):
+
+    def __init__(self):
+        self.cached_cells = []
+        self._expired = True
+
+    def expired(self):
+        return self._expired == True
+
+    def close(self):
+        self.cached_cells = []
+        self._expired = True
+
+    def store(self, data, cell_start=CellIndex(col=1, row=1)):
+        self.cache_cells = [
+            [
+                gspread.Cell(
+                    row=nrow + cell_start.row,
+                    col=ncol + cell_start.col,
+                    value=val
+                ) for ncol,val in enumerate(row)
+            ] for nrow,row in enumerate(data)]
+        self._expired = False
+
 class GoogleSheets(object):
 
     class AlreadyExists (Exception):
@@ -66,12 +92,19 @@ class GoogleSheets(object):
     class NotFound (Exception):
         pass
 
+    class InitError (Exception):
+        pass
+
     """
     run_mode: service for service account, user for oauth (require user interaction)
     if not set at object creation like gs = GoogleSheets(run_mode='service')
     will take the valie from GOOGLESHEETS_RUN_MODE env variable and default to 'service'
     """
-    def __init__(self, run_mode=''):
+    def __init__(self, run_mode='',
+                 scopes=DEFAULT_SCOPES,
+                 flow=local_server_flow,
+                 credentials_filename=DEFAULT_CREDENTIALS_FILENAME,
+                 authorized_user_filename=DEFAULT_AUTHORIZED_USER_FILENAME):
         assert run_mode in (None, '', 'service', 'user'), "run_mode set and not in 'service' or 'user'"
         self.run_mode = run_mode if run_mode else getenv('GOOGLESHEETS_RUN_MODE', 'service')
         self.spreadsheet_cursor = None
@@ -82,11 +115,46 @@ class GoogleSheets(object):
             self.gc = gspread.service_account()
             self.gc = ClientRetry(auth=self.gc.auth)
         else:
-            self.gc = gspread.oauth()
-            self.gc = ClientRetry(auth=gspread.auth.load_credentials())
-
+            for i in [1, 2]:
+                self.gc = gspread.oauth(
+                    scopes=scopes,
+                    flow=flow,
+                    credentials_filename=credentials_filename,
+                    authorized_user_filename=authorized_user_filename
+                )
+                self.gc = ClientRetry(auth=gspread.auth.load_credentials())
+                # validate grant by attempting a request
+                try:
+                    tmp = self.gc.list_spreadsheet_files(title='')
+                except Exception as e:
+                    logger.error (dir(e))
+                    if 'RefreshError' in type(e).__name__:
+                        if path.exists (authorized_user_filename):
+                            unlink (authorized_user_filename)
+                            continue
+                    else:
+                        raise e
+                else:
+                    break
+            else:
+                raise self.InitError from None
         self.client_ext = ClientRetry(self.gc.auth, self.gc.session)
         self.placeholder = []
+        self.data_cache = DataCache()
+
+    """
+    id
+    """
+    def spreadsheet_id(self):
+        if self.spreadsheet_cursor:
+            return self.spreadsheet_cursor.id
+
+    """
+    title
+    """
+    def spreadsheet_title(self):
+        if self.spreadsheet_cursor:
+            return self.spreadsheet_cursor.title
 
     """
     Creates a new spreadsheet.
@@ -95,7 +163,7 @@ class GoogleSheets(object):
         if self.spreadsheet_cursor is None:
             self.spreadsheet_cursor = SpreadsheetRetry(self.gc.create(title=title, folder_id=folder_id))
             logger.info ("create: {}".format(self.spreadsheet_cursor))
-            return self.spreadsheet_cursor
+
 
     """
     Deletes a spreadsheet.
@@ -103,13 +171,15 @@ class GoogleSheets(object):
     def delete_spreadsheet(self):
         logger.info ("delete: {}".format(self.spreadsheet_cursor))
         if self.spreadsheet_cursor is not None:
-            self.gc.del_spreadsheet(self.spreadsheet_cursor.id)
-            self.spreadsheet_cursor = None
-            self.worksheet_cursor = None
-            self.cell_current_position = (1, 1)
-            self.spreadsheet_revision = None
+            try:
+                self.gc.del_spreadsheet(self.spreadsheet_cursor.id)
+            except Exception as e:
+                logger.error ("delete_spreadsheet: {}".format(e))
+                raise e
+            else:
+                self.close()
         else:
-            raise NotFound ("delete spreadsheet not open/created")
+            raise self.NotFound ("delete spreadsheet not open/created")
 
     """
     Give permission to a spreadsheet
@@ -156,27 +226,50 @@ class GoogleSheets(object):
             return self.spreadsheet_cursor.list_permissions()
 
     """
-    Open a spreadsheet
+    """
+    def is_open(self):
+        return self.spreadsheet_cursor is not None
+
+    """
+    may be used to open an other spreadsheet reusing an existing GoogleSheet instance
+    """
+    def close(self):
+        self.spreadsheet_cursor = None
+        self.worksheet_cursor = None
+        self.cell_current_position = (1, 1)
+        self.spreadsheet_revision = None
+        self.data_cache.close()
+
+    """
+    Open a spreadsheet try in order 'url', id and then title
+    if nay tab_name, tab_position, tab_id is set open the worsheet trying first
+    'tab_id', 'tab_name', 'tab_position' and then tab '0'
     """
     @retry(tries=15, delay=5, backoff=2, except_retry=[error_quota_req])
     def open (self, title=None, url=None, key=None, tab_name=None, tab_position=None, tab_id=None):
         if self.spreadsheet_cursor is None:
             assert any ([title, url, key]), "opening a spreadsheet require a title, an url or a key"
-            if title:
+            for i in ['url', 'key', 'title']:
                 try:
-                    self.spreadsheet_cursor = SpreadsheetRetry(self.gc.open(title))
+                    if i == 'url' and url:
+                        self.spreadsheet_cursor = SpreadsheetRetry(self.gc.open_by_url(url))
+                    if i == 'key' and key:
+                        self.spreadsheet_cursor = SpreadsheetRetry(self.gc.open_by_key(key))
+                    if i == 'title' and title:
+                        self.spreadsheet_cursor = SpreadsheetRetry(self.gc.open(title))
+                    if self.spreadsheet_cursor is None:
+                        continue
+                    else:
+                        break
                 except exceptions.SpreadsheetNotFound as e:
-                    raise self.NotFound from e
-            if url:
-                try:
-                    self.spreadsheet_cursor = SpreadsheetRetry(self.gc.open_by_url(url))
-                except exceptions.SpreadsheetNotFound as e:
-                    raise self.NotFound from e
-            if key:
-                try:
-                    self.spreadsheet_cursor = SpreadsheetRetry(self.gc.open_by_key(key))
-                except exceptions.SpreadsheetNotFound as e:
-                    raise self.NotFound from e
+                    continue
+                except Exception as e:
+                    logger.error (e)
+                    raise
+                else:
+                    break
+
+            if not self.spreadsheet_cursor: raise self.NotFound
             logger.info ("open spreadsheet: {}".format(self.spreadsheet_cursor))
             try:
                 self.spreadsheet_revision = self.client_ext.revision_last(
@@ -190,41 +283,54 @@ class GoogleSheets(object):
             else:
                 pass
                 # logger.info ("spreadsheet_revision: {}".format(self.spreadsheet_revision))
-
-            if not any ([tab_name, tab_position, tab_id]): return self.spreadsheet_cursor
+            self.data_cache.close()
+            if not any ([tab_name, tab_position, tab_id]): return
 
         self.worksheet_cursor = None
-        if tab_name:
-            logger.debug ("open tab name {}".format(tab_name))
-            self.worksheet_cursor = [
-                w for w in [
-                    w for w in self.worksheets()
-                ] if w.title.strip().lower() in [
-                    n.strip().lower() for n in [tab_name]]]
-            assert self.worksheet_cursor, "error in open tab name: {}".format(tab_name)
-            self.worksheet_cursor = self.worksheet_cursor[0]
-        elif tab_id:
-            tab_id = int (tab_id)
-            logger.debug ("open tab id {}".format(tab_id))
-            self.worksheet_cursor = [
-                w for w in [
-                    w for w in self.worksheets()
-                ] if w.id in [tab_id]]
-            assert self.worksheet_cursor, "error in open tab id: {}".format([tab_id])
-            self.worksheet_cursor = self.worksheet_cursor[0]
-        elif tab_position:
-            logger.debug ("open tab pos {}".format(tab_position))
-            self.worksheet_cursor = [
-                (i,w) for (i,w) in
-                enumerate(self.worksheets()) if i == int(tab_position)]
-            assert self.worksheet_cursor, "error in open tab position: {}".format(tab_position)
+        for i in ['tab_id', 'tab_name', 'tab_position']:
+            try:
+                if i == 'tab_name' and tab_name:
+                    logger.debug ("open tab name {}".format(tab_name))
+                    self.worksheet_cursor = [
+                        w for w in [
+                            w for w in self.worksheets()
+                        ] if w.title.strip().lower() in [
+                            n.strip().lower() for n in [tab_name]]]
+                    assert self.worksheet_cursor, "error in open tab name: {}".format(tab_name)
+                    self.worksheet_cursor = self.worksheet_cursor[0]
+                if i == 'tab_id' and tab_id:
+                    tab_id = int (tab_id)
+                    logger.debug ("open tab id {}".format(tab_id))
+                    self.worksheet_cursor = [
+                        w for w in [
+                            w for w in self.worksheets()
+                        ] if w.id in [tab_id]]
+                    assert self.worksheet_cursor, "error in open tab id: {}".format([tab_id])
+                    self.worksheet_cursor = self.worksheet_cursor[0]
+                if i == 'tab_position' and tab_position:
+                    logger.debug ("open tab pos {}".format(tab_position))
+                    self.worksheet_cursor = [
+                        (i,w) for (i,w) in
+                        enumerate(self.worksheets()) if i == int(tab_position)]
+                    assert self.worksheet_cursor, "error in open tab position: {}".format(tab_position)
+                if self.worksheet_cursor is None:
+                    continue
+                else:
+                    break
+            except AssertionError as e:
+                logger.error (e)
+                continue
+            except Exception as e:
+                logger.error (e)
+            else:
+                break
         else:
-            logger.debug ("open tab pos {}".format(0))
+            logger.debug ("warning open tab pos {}".format(0))
             self.worksheet_cursor = [
                 (i,w) for (i,w) in
                 enumerate(self.worksheets()) if i == int(0)]
             assert self.worksheet_cursor, "error in open tab position: {}".format(0)
-        return self.worksheet_cursor
+        self.data_cache.close()
 
     """
     return a list with all the fields for each revision available
@@ -241,24 +347,38 @@ class GoogleSheets(object):
         result = self.client_ext.revision_list_mtime(spreadsheet_id=self.spreadsheet_cursor.id)
         return result
 
+    """
+    given one of the supported format extension,
+    return a mime type suitable for the file _export function
+    """
+    def ext2mime (self, extension):
+        mimes = {
+            'csv': 'text/csv',
+            'ods': 'application/vnd.oasis.opendocument.spreadsheet',
+            'pdf': 'application/pdf',
+            'tsv': 'text/tab-separated-values',
+            'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
+            'zip': 'application/zip',
+        }
+        logger.debug ("ext2mime: {}".format(extension.lower()))
+        if extension.lower() in mimes:
+            return mimes[extension.lower()]
+        else:
+            raise self.NotFound ("no mime map for {}".format(extension))
+
     """
     Export a specific revision, or last modified if revision_id=None
     usage:
        with open ('/tmp/demo-revision-1.pdf', 'wb') as fd:
         gs.file_export (fd, revision_id=1, mime_type='application/pdf')
-
-    mime_type                                                        format
-
-    application/pdf                                                   pdf
-    application/vnd.oasis.opendocument.spreadsheet                    ods
-    application/vnd.openxmlformats-officedocument.spreadsheetml.sheet xlsx
-    application/x-vnd.oasis.opendocument.spreadsheet                  ods
-    application/zip                                                   zip
-    text/csv                                                          csv
-    text/tab-separated-values                                         tsv
     """
-    def file_export(self, fd, revision_id='head',
-                    mime_type='application/x-vnd.oasis.opendocument.spreadsheet'):
+    def file_export(self, fd, revision_id='head', mime_type=None, extension=None):
+        if not mime_type and extension:
+            mime_type = self.ext2mime(extension)
+        elif not mime_type and not extension:
+            if fd.name:
+                mime_type = self.ext2mime(fd.name.rpartition('.')[-1])
+        assert mime_type, "unknow mime_type"
         result = self.client_ext.file_export(
             fd, spreadsheet_id=self.spreadsheet_cursor.id, revision_id=revision_id, mime_type=mime_type)
         return result
@@ -267,8 +387,10 @@ class GoogleSheets(object):
     upload content of file descripor fd on success return new GoogleSheets if return_object == True
     else a file ressource id like '1qhNTwrt6BGcOX3c3DqaINMLeIsc-ceHJ'
     """
-    def file_upload(self, fd, title='', mime_type='application/x-vnd.oasis.opendocument.spreadsheet',
+    def file_upload(self, fd, title='', mime_type=None, extension=None,
                     return_object=True):
+        if extension:
+            mime_type= self.ext2mime(extension)
         new_id = self.client_ext.file_upload(fd, title=title, mime_type=mime_type)
         if return_object:
             new_object = GoogleSheets(run_mode=self.run_mode)
@@ -299,13 +421,41 @@ class GoogleSheets(object):
             """ ori clean up"""
             for w in self.worksheets():
                 if w.id in wid_ori:
-                    self.spreadsheet_cursor.del_worksheet(w)
+                    try:
+                        self.spreadsheet_cursor.del_worksheet(w)
+                    except Exception as e:
+                        err_name = "{}.{}".format(e.__class__.__module__ ,  e.__class__.__name__)
+                        err_code = None
+                        err_code = [i['code'] for i in e.args if 'code' in i]
+                        err_code = int(err_code[0]) if err_code else None
+                        logger.debug ("exception handling {} {}".format(err_name, err_code))
+                        if err_code == 400:
+                            # handle protected cell or object
+                            logger.warning ("L434: {}".format(e))
+                        else:
+                            raise e
             """ cp from src to self """
             for w in src.worksheets():
+                logger.info ("COPY {}.copy_to({})".format(w, self.spreadsheet_cursor.id))
                 w.copy_to(self.spreadsheet_cursor.id)
             """ copy prepand with 'Copy of' """
             for w in self.worksheets():
-                w.update_title(w.title.replace("Copy of", "").strip())
+                try:
+                    logger.info (
+                        "UPDATE {}.update_title({})".format(w, w.title.replace("Copy of", "").strip())
+                    )
+                    w.update_title(w.title.replace("Copy of", "").strip())
+                except Exception as e:
+                    err_name = "{}.{}".format(e.__class__.__module__ ,  e.__class__.__name__)
+                    err_code = None
+                    err_code = [i['code'] for i in e.args if 'code' in i]
+                    err_code = int(err_code[0]) if err_code else None
+                    logger.debug ("exception handling {} {}".format(err_name, err_code))
+                    if err_code == 400:
+                        # handle protected cell or object
+                        logger.warning ("L456: {}".format(e))
+                    else:
+                        raise e
             self.open(tab_name="{}".format(self))
             self.delete_worksheet()
         except Exception as e:
@@ -347,6 +497,7 @@ class GoogleSheets(object):
                     self.worksheet_cursor = self.worksheet_cursor[0]
             else: raise e
         else:
+            self.data_cache.close()
             logger.info("create {}".format(self.worksheet_cursor))
 
     """
@@ -372,6 +523,7 @@ class GoogleSheets(object):
         assert self.worksheet_cursor is not None, "no active worksheet to delete"
         logger.info ("delete {}".format(self.worksheet_cursor))
         self.worksheet_cursor = self.spreadsheet_cursor.del_worksheet(self.worksheet_cursor)
+        self.data_cache.close()
         self.worksheet_cursor = None
 
     """
@@ -380,31 +532,55 @@ class GoogleSheets(object):
     def resize(self, cols=None, rows=None):
         assert self.worksheet_cursor is not None, "no active worksheet to resize"
         self.worksheet_cursor.resize(cols=cols, rows=rows)
+        self.data_cache.close()
         logger.info ("{} col_count={} row_count={}".format(
             self.worksheet_cursor, self.worksheet_cursor.col_count, self.worksheet_cursor.row_count))
 
+    """
+    clear all cached data
+    """
+    def close_cache (self):
+        self.data_cache.close()
+
     """
     lookup match in search_direction  X (col) or Y (row)
     return a list of GridIndex (start.col, start.row, end.col, end.row) if match else None
     the result list is sorted with the longest at the end
 
-    as the funtion us regexpr it may be needed to validate by fetching the data
+    as the funtion use regexpr it may be needed to validate by fetching the data
     """
     def lookup_match (self, match=[], search_direction='col', default_regex=r"\b({})\b"):
         assert self.worksheet_cursor, "worksheet not open"
-        match_list = [
-            compile(default_regex.format(m), IGNORECASE) for m in match]
-        logger.info ("match_list: {}".format (match_list))
+
+        if self.data_cache.expired():
+            data = self.get_values()
+            self.data_cache.store (data)
+
+        rs = ""
+        for i in match[:-1]:
+            rs += default_regex.format(i) + "|" if i else ''
+        rs += default_regex.format(match[-1]) if match and match[-1] else '(^$)'
+        logger.info ("r: {}".format(rs))
+        rc = compile(rs, IGNORECASE)
+
         cell_find_list = []
-        for match in match_list:
-            cell_find_list += self.worksheet_cursor.findall(match)
-        logger.debug("findall: {}".format(cell_find_list))
+        for i in self.data_cache.cache_cells:
+            for j in i:
+                if rc.search (j.value):
+                    cell_find_list.append (j)
+                elif j.value == '':
+                    if cell_find_list and search_direction in ('col', 'x') and (
+                            cell_find_list[-1].row == j.row and cell_find_list[-1].col == j.col - 1):
+                        cell_find_list.append (j)
+                    elif cell_find_list and search_direction in ('row', 'y') and (
+                            cell_find_list[-1].col == j.col and cell_find_list[-1].row == j.row - 1):
+                        cell_find_list.append (j)
+
+        logger.debug("find: {}".format(cell_find_list))
         if search_direction.lower() in ('col', 'x'):
             cell_find_list.sort(key=lambda x: (int(x.row), int(x.col)), reverse=False)
         else:
             cell_find_list.sort(key=lambda x: (int(x.col), int(x.row)), reverse=False)
-        # for l in cell_find_list:
-        #     logger.info ("sorted: {}".format(l))
 
         result=[]
         ridx=GridIndex()
@@ -459,7 +635,7 @@ class GoogleSheets(object):
             range_name = "{}:{}".format(s, e)
         else:
             range_name = None
-        result = self.worksheet_cursor.get_values(range_name=range_name)
+        result = self.worksheet_cursor.get_values(range_name=range_name, **kwargs)
         return result
 
     """
@@ -468,6 +644,7 @@ class GoogleSheets(object):
     def delete_cols(self, start_index, end_index=None):
         assert self.worksheet_cursor, "worksheet not open"
         self.worksheet_cursor.delete_columns(start_index, end_index=end_index)
+        self.data_cache.close()
 
     """
     delete rows from start to end
@@ -475,6 +652,7 @@ class GoogleSheets(object):
     def delete_rows(self, start_index, end_index=None):
         assert self.worksheet_cursor, "worksheet not open"
         self.worksheet_cursor.delete_rows(start_index, end_index=end_index)
+        self.data_cache.close()
 
     """
     update a single cell value
@@ -482,6 +660,7 @@ class GoogleSheets(object):
     def update_cell(self, col, row, value):
         assert self.worksheet_cursor, "worksheet not open"
         self.worksheet_cursor.update_cell(col=col, row=row, value=value)
+        self.data_cache.close()
 
     """
     Clears multiple ranges in one API call
@@ -510,6 +689,7 @@ class GoogleSheets(object):
             else:
                 raise ValueError ("clear idx {}".format(grid_idx))
         result = self.worksheet_cursor.batch_clear(ranges=range_list, **kwargs)
+        self.data_cache.close()
         return result
 
     """
@@ -528,7 +708,8 @@ class GoogleSheets(object):
                     values=[[1, 8, 6, 4, 2]])
     KO use transpose=True or pass the data as : [[1], [8], [6], [4], [2]]
     """
-    def update_cells(self, cells_index, values, transpose=False, **kwargs):
+    def update_cells(self, cells_index, values, transpose=False,
+                     value_input_option=ValueInputOption.user_entered,**kwargs):
         assert self.worksheet_cursor, "worksheet not open"
         start_col = 0
         start_row = 0
@@ -556,9 +737,37 @@ class GoogleSheets(object):
             for v in c:
                 cell_list[idx].value = v
                 idx += 1
-        result = self.worksheet_cursor.update_cells(cell_list=cell_list, **kwargs)
+        result = self.worksheet_cursor.update_cells(
+            cell_list=cell_list, value_input_option=value_input_option, **kwargs)
+        self.data_cache.close()
         return result
 
+    """
+    refresh_ref
+    try to refresh the reference in a spreadsheet by overwriting the same formula
+    return the number of reference still unresoved in the spreadsheet
+    """
+    def refresh_ref (self):
+        ref_count = 0
+        t = self.spreadsheet_title()
+        for w in self.worksheets():
+            logger.info ("s {} w {}".format(t, w))
+            self.open(tab_id=w.id)
+            match_location = self.lookup_match (match=['#REF!'], default_regex=r"^{}$")
+            for m in match_location:
+                logger.info ("refresh {}: {}".format(w, m))
+                v = self.get_values(m, value_render_option=ValueRenderOption.formula)
+                try:
+                    self.update_cells (m, v, value_input_option=ValueInputOption.user_entered)
+                except Exception as e:
+                    logger.warning ("refresh_ref {}".format(e))
+            match_location = self.lookup_match (match=['#REF!'], default_regex=r"^{}$")
+            for m in match_location:
+                ref_count += 1
+                logger.warning ("refresh ref unresolved {} {}".format(w, m))
+        if ref_count > 0: logger.info ("refresh_ref {} unresolved ref in {}".format(ref_count, t))
+        return ref_count
+
 
     """
     Cell Formatting
